#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#include <util/delay.h>

#include "video.h"
#include "tilp.h"
#include "ti_real.h"

#define TRIGGER_PORT PORTB // The port the trigger button is connected to.
#define TRIGGER_DDR   DDRB // The direction register the trigger button is connected to.
#define TRIGGER_PIN   PINB // The pin register the trigger button is connected to.
#define TRIGGER          3 // The bit index within the port the trigger button is connected to.

#define ACTIVITY_PORT PORTB // The port the activity LED is connected to.
#define ACTIVITY_DDR   DDRB // The direction register the activity LED is connected to.
#define ACTIVITY_PIN   PINB // The pin register the activity LED is connected to.
#define ACTIVITY          4 // The bit index within the port the activity LED is connected to.

uint8_t tilp_screenshot_request[] PROGMEM = { 0x02, 0x6D, 0x00, 0x00 };
uint8_t tilp_variable_request[] PROGMEM = { 0x73, 0xA2, 0x0D, 0x00 };
uint8_t tilp_variable_clear_to_send[] PROGMEM = { 0x73, 0x09, 0x00, 0x00 };
uint8_t tilp_packet_header[4];
uint8_t tilp_variable_transfer[13];

#define EEPROM_ID 0x54564445
#define EEPROM_OFFSET_ID 0
#define EEPROM_OFFSET_VIDEO_STANDARD 4
#define EEPROM_OFFSET_PAL_X 5
#define EEPROM_OFFSET_PAL_Y 6
#define EEPROM_OFFSET_NTSC_X 8
#define EEPROM_OFFSET_NTSC_Y 9

void set_variable(char variable_name) {
	
	int16_t value = ti_real_get(tilp_variable_transfer);
	
	switch (variable_name) {
		case 'M':
			switch (value) {
				case 50:
					video_init(VIDEO_STANDARD_PAL);
					break;
				case 60:
					video_init(VIDEO_STANDARD_NTSC);
					break;
			}
			eeprom_update_byte((uint8_t*)EEPROM_OFFSET_VIDEO_STANDARD, video_standard);
			break;
		case 'X':
			switch (video_standard) {
				case VIDEO_STANDARD_PAL:
					if (value < VIDEO_PAL_X_OFFSET_MIN) value = VIDEO_PAL_X_OFFSET_MIN;
					if (value > VIDEO_PAL_X_OFFSET_MAX) value = VIDEO_PAL_X_OFFSET_MAX;
					video_pal_x_offset = value;
					eeprom_update_byte((uint8_t*)EEPROM_OFFSET_PAL_X, video_pal_x_offset);
					break;
				case VIDEO_STANDARD_NTSC:
					if (value < VIDEO_NTSC_X_OFFSET_MIN) value = VIDEO_NTSC_X_OFFSET_MIN;
					if (value > VIDEO_NTSC_X_OFFSET_MAX) value = VIDEO_NTSC_X_OFFSET_MAX;
					video_ntsc_x_offset = value;
					eeprom_update_byte((uint8_t*)EEPROM_OFFSET_NTSC_X, video_ntsc_x_offset);
					break;
			}
			break;
		case 'Y':
			value = -value;
			switch (video_standard) {
				case VIDEO_STANDARD_PAL:
					if (value < VIDEO_PAL_Y_OFFSET_MIN) value = VIDEO_PAL_Y_OFFSET_MIN;
					if (value > VIDEO_PAL_Y_OFFSET_MAX) value = VIDEO_PAL_Y_OFFSET_MAX;
					video_pal_y_offset = value;
					eeprom_update_word((uint16_t*)EEPROM_OFFSET_PAL_Y, video_pal_y_offset);
					break;
				case VIDEO_STANDARD_NTSC:
					if (value < VIDEO_NTSC_Y_OFFSET_MIN) value = VIDEO_NTSC_Y_OFFSET_MIN;
					if (value > VIDEO_NTSC_Y_OFFSET_MAX) value = VIDEO_NTSC_Y_OFFSET_MAX;
					video_ntsc_y_offset = value;
					eeprom_update_word((uint16_t*)EEPROM_OFFSET_NTSC_Y, video_ntsc_y_offset);
					break;
			}
			break;
	}
}

int main(void) {

	// Initilise the trigger button.
	TRIGGER_PORT |= _BV(TRIGGER);
	
	// Initialise the activity LED.
	ACTIVITY_DDR |= _BV(ACTIVITY);

	// Brief delay to give everything time to wake up.
	_delay_ms(50);

	// Are we resetting the settings?
	uint8_t reset_settings = 0;
	for (uint16_t timer = 0; (timer < 2000) && (reset_settings = !(TRIGGER_PIN & _BV(TRIGGER))); ++timer) {
		_delay_ms(1);
	}
	
	if (reset_settings || eeprom_read_dword((const uint32_t*)EEPROM_OFFSET_ID) != EEPROM_ID) {
		// EEPROM does not contain values, so update manually.
		eeprom_update_byte((uint8_t*)EEPROM_OFFSET_VIDEO_STANDARD, VIDEO_STANDARD_NTSC);
		eeprom_update_byte((uint8_t*)EEPROM_OFFSET_PAL_X, video_pal_x_offset);
		eeprom_update_word((uint16_t*)EEPROM_OFFSET_PAL_Y, video_pal_y_offset);
		eeprom_update_byte((uint8_t*)EEPROM_OFFSET_NTSC_X, video_ntsc_x_offset);
		eeprom_update_word((uint16_t*)EEPROM_OFFSET_NTSC_Y, video_ntsc_y_offset);
		eeprom_update_dword((uint32_t*)EEPROM_OFFSET_ID, EEPROM_ID);
	} else {	
		// Load values from EEPROM.
		video_standard = eeprom_read_byte((const uint8_t*)EEPROM_OFFSET_VIDEO_STANDARD);
		video_pal_x_offset = eeprom_read_byte((const uint8_t*)EEPROM_OFFSET_PAL_X);
		video_pal_y_offset = eeprom_read_word((const uint16_t*)EEPROM_OFFSET_PAL_Y);
		video_ntsc_x_offset = eeprom_read_byte((const uint8_t*)EEPROM_OFFSET_NTSC_X);
		video_ntsc_y_offset = eeprom_read_word((const uint16_t*)EEPROM_OFFSET_NTSC_Y);
	}
	
	// Initialise libraries.
	video_init(video_standard);
	tilp_init();
	
	// Enable interrupts.
	sei();
	
	// Wait for the trigger to be released.
	while (!(TRIGGER_PIN & _BV(TRIGGER))) {
		_delay_ms(10);
	}
	
	uint8_t tilp_byte;
	
	for (;;) {
		
		if (!(TRIGGER_PIN & _BV(TRIGGER))) {
			
			// The trigger button has been pressed.
			
			ACTIVITY_PORT |= _BV(ACTIVITY);
			
			// Screenshot request.
			if (tilp_write_block_P(tilp_screenshot_request, sizeof(tilp_screenshot_request)) == TILP_OK &&
				tilp_read_acknowledgement(&tilp_byte) == TILP_OK
			) {
			
				// Read the screenshot.
				if (tilp_read_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK && // Read the header.
					tilp_packet_header[1] == 0x15 && // Check it's a screenshot.
					tilp_packet_header[2] == 0x00 && tilp_packet_header[3] == 0x03 && // Check it's 768 bytes (0x300) long.
					tilp_read_block_and_checksum(video_buffer, 768) == TILP_OK // Read the screenshot data.
				) {
					tilp_write_acknowledgement(0x02);
				}				
			}
			
			ACTIVITY_PORT &= (uint8_t)~_BV(ACTIVITY);
			
		} else if ((TILP_PIN & (_BV(TILP_RED) | _BV(TILP_WHITE))) != (_BV(TILP_RED) | _BV(TILP_WHITE))) {
			ACTIVITY_PORT |= _BV(ACTIVITY);
			
			// Something is trying to send us data.
			if (tilp_read_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK) {
				switch (tilp_packet_header[1]) {
					case 0x68: {
						// Check ready.
						tilp_write_acknowledgement(0x73);
					} break;
					
					case 0x6D: {
						
						// Take screenshot (SLP).
						if (tilp_write_acknowledgement(0x73) == TILP_OK &&
							tilp_write_byte(tilp_packet_header[0]) == TILP_OK &&
							tilp_write_byte(0x15) == TILP_OK &&
							tilp_write_byte(0x00) == TILP_OK &&
							tilp_write_byte(0x03) == TILP_OK &&
							tilp_write_block_and_checksum(video_buffer, 768) == TILP_OK
						) {
							tilp_read_acknowledgement(&tilp_byte);
						}
						
					} break;
					
					case 0x06: {						
						// Send variable (Manual).
						if (tilp_packet_header[2] == 0x0D && tilp_packet_header[3] == 0x00) {
							if (tilp_read_block_and_checksum(tilp_variable_transfer, 13) == TILP_OK && tilp_write_acknowledgement(0x73) == TILP_OK) {
								switch (tilp_variable_transfer[0]) {
									case 0x09: { // Real number.
										char variable_name = tilp_variable_transfer[3];
										tilp_packet_header[1] = 0x09;
										tilp_packet_header[2] = 0x00;
										tilp_packet_header[3] = 0x00;
										if (tilp_write_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK &&
											tilp_read_acknowledgement(&tilp_byte) == TILP_OK &&
											tilp_read_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK &&
											tilp_packet_header[1] == 0x15 && tilp_packet_header[2] == 0x09 && tilp_packet_header[3] == 0x00 &&
											tilp_read_block_and_checksum(tilp_variable_transfer, 9) == TILP_OK &&
											tilp_write_acknowledgement(0x73) == TILP_OK
										) {
											set_variable(variable_name);
										}
									} break;
									default: {
										tilp_packet_header[1] = 0x36;
										tilp_packet_header[2] = 0x01;
										tilp_packet_header[3] = 0x00;
										if (tilp_write_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK) {
											tilp_write_byte(0x02);
											tilp_write_byte(0x02);
											tilp_write_byte(0x00);
											tilp_read_acknowledgement(&tilp_byte);
										}
									} break;
								}
							}
						}
						
					} break;
					
					case 0xA2: { // Get()
						if (tilp_packet_header[2] == 0x0B && tilp_packet_header[3] == 0x00) {
							if (tilp_read_block_and_checksum(tilp_variable_transfer, 11) == TILP_OK && tilp_write_acknowledgement(0x83) == TILP_OK) {
								char variable_name = tilp_variable_transfer[3];
								tilp_packet_header[0] = 0x83;
								tilp_packet_header[1] = 0x06;
								tilp_variable_transfer[0] = 0x09;
								if (tilp_write_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK &&
									tilp_write_block_and_checksum(tilp_variable_transfer, 11) == TILP_OK &&
									tilp_read_acknowledgement(&tilp_byte) == TILP_OK &&
									tilp_read_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK && tilp_packet_header[1] == 0x09 &&
									tilp_write_acknowledgement(0x73) == TILP_OK &&
									(tilp_packet_header[1] = 0x15, tilp_packet_header[2] = 0x09, tilp_packet_header[3] = 0x00, tilp_write_block(tilp_packet_header, sizeof(tilp_packet_header))) == TILP_OK
								) {
									int16_t value = 0;
									switch (variable_name) {
										case 'M':
											switch (video_standard) {
												case VIDEO_STANDARD_PAL:
													value = 50;
													break;
												case VIDEO_STANDARD_NTSC:
													value = 60;
													break;
											}
											break;
										case 'X':
											switch (video_standard) {
												case VIDEO_STANDARD_PAL:
													value = video_pal_x_offset;
													break;
												case VIDEO_STANDARD_NTSC:
													value = video_ntsc_x_offset;
													break;
											}
											break;
										case 'Y':
											switch (video_standard) {
												case VIDEO_STANDARD_PAL:
													value = -video_pal_y_offset;
													break;
												case VIDEO_STANDARD_NTSC:
													value = -video_ntsc_y_offset;
													break;
											}
											break;
										default:
											break;
									}
									
									ti_real_set(tilp_variable_transfer, value);
									
									if (tilp_write_block_and_checksum(tilp_variable_transfer, 9) == TILP_OK) {
										tilp_read_acknowledgement(&tilp_byte);
									}
								}
							} break;
						}
					} break;
					
					case 0xC9: { // Send()
						if (tilp_packet_header[2] == 0x0B && tilp_packet_header[3] == 0x00) {
							if (tilp_read_block_and_checksum(tilp_variable_transfer, 11) == TILP_OK && tilp_write_acknowledgement(0x73) == TILP_OK) {
								// I think that as we just acknowledged the variable request as a TI-83+ (rather than a CBL/CBR) the calculator is satisfied.
								// However, we can now start a regular (documented) silent variable transfer from the calculator.
								char variable_name = tilp_variable_transfer[3];
								tilp_variable_transfer[0] = 0x09;
								tilp_variable_transfer[11] = 0x00;
								tilp_variable_transfer[12] = 0x00;
								if (tilp_write_block_P(tilp_variable_request, sizeof(tilp_variable_request)) == TILP_OK &&
									tilp_write_block_and_checksum(tilp_variable_transfer, 13) == TILP_OK &&
									tilp_read_acknowledgement(&tilp_byte) == TILP_OK &&
									tilp_read_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK && tilp_packet_header[1] == 0x06 &&
									tilp_read_block_and_checksum(tilp_variable_transfer, 13) == TILP_OK &&
									tilp_write_acknowledgement(0x73) == TILP_OK &&
									(tilp_write_block_P(tilp_variable_clear_to_send, sizeof(tilp_variable_clear_to_send))) == TILP_OK &&
									tilp_read_acknowledgement(&tilp_byte) == TILP_OK &&
									tilp_read_block(tilp_packet_header, sizeof(tilp_packet_header)) == TILP_OK && tilp_packet_header[1] == 0x15 &&
									tilp_read_block_and_checksum(tilp_variable_transfer, 9) == TILP_OK &&
									tilp_write_acknowledgement(0x73) == TILP_OK
								) {
									set_variable(variable_name);
								}
							}
						}
					} break;
					
					case 0x92: { // No more variables.
						tilp_write_acknowledgement(0x73);
					} break;
					
				}
			}
			ACTIVITY_PORT &= (uint8_t)~_BV(ACTIVITY);
		}
	}
}