#include <stdint.h>

#include <avr/io.h>
#include <avr/pgmspace.h>

#include "tilp.h"

// Initialises the TI link protocol routines.
void tilp_init(void) {
	// Float both pins.
	TILP_LINE_HIGH(TILP_RED);
	TILP_LINE_HIGH(TILP_WHITE);
	
	// Timer 0 overflows after (1/20MHz)*(1024*256)=13.11ms
	TCCR0B = _BV(CS02) | _BV(CS00); // clkIO/1024 (From prescaler).
}

// Used to time time-out events.
uint8_t timed_out_counter;

// Determines whether the link has timed out or not. Needs to be called periodically.
uint8_t not_timed_out(void) {
	if (TIFR0 & _BV(TOV0)) {
		TIFR0 |= _BV(TOV0);
		if (timed_out_counter > 0) {
			--timed_out_counter;
		}
	}
	return timed_out_counter;
}

// Write a single byte using the TI link protocol.
enum tilp_result_t tilp_write_byte(uint8_t value) {
	TCNT0 = 0; timed_out_counter = TILP_TIME_OUT_TIME; TIFR0 |= _BV(TOV0);
	for (uint8_t bit = 0; bit < 8 && not_timed_out(); ++bit) {
		if (value & 1) {
			TILP_LINE_LOW(TILP_WHITE);
			while (TILP_LINE(TILP_RED) && not_timed_out());
			TILP_LINE_HIGH(TILP_WHITE);
			while (!TILP_LINE(TILP_RED) && not_timed_out());
		} else {
			TILP_LINE_LOW(TILP_RED);
			while (TILP_LINE(TILP_WHITE) && not_timed_out());
			TILP_LINE_HIGH(TILP_RED);
			while (!TILP_LINE(TILP_WHITE) && not_timed_out());
		}
		value >>= 1;
	}
	return not_timed_out() ? TILP_OK : TILP_TIME_OUT;
}

// Read a single byte using the TI link protocol.
enum tilp_result_t tilp_read_byte(uint8_t* value) {
	TCNT0 = 0; timed_out_counter = TILP_TIME_OUT_TIME; TIFR0 |= _BV(TOV0);
	*value = 0;
	for (uint8_t bit = 0; bit < 8 && not_timed_out(); ++bit) {
		*value >>= 1;
		uint8_t line_state = ~(_BV(TILP_RED) | _BV(TILP_WHITE));
		while (((line_state == (_BV(TILP_RED) | _BV(TILP_WHITE))) || (line_state != (TILP_PIN & (_BV(TILP_RED) | _BV(TILP_WHITE))))) && not_timed_out()) {
			line_state = TILP_PIN & (_BV(TILP_RED) | _BV(TILP_WHITE));
		}
		if (line_state & _BV(TILP_RED)) {
			*value |= 0x80;
			TILP_LINE_LOW(TILP_RED);
			while (!TILP_LINE(TILP_WHITE) && not_timed_out());
			TILP_LINE_HIGH(TILP_RED);
		} else {
			TILP_LINE_LOW(TILP_WHITE);
			while (!TILP_LINE(TILP_RED) && not_timed_out());
			TILP_LINE_HIGH(TILP_WHITE);
		}
	}
	return not_timed_out() ? TILP_OK : TILP_TIME_OUT;
}

// Write a block of values using the TI link protocol.
enum tilp_result_t tilp_write_block(uint8_t* block, uint16_t block_length) {
	enum tilp_result_t result = TILP_OK;
	while (block_length-- && result == TILP_OK) {
		result = tilp_write_byte(*block++);
	}
	return result;
}

// Write a block of values and a checksum using the TI link protocol.
enum tilp_result_t tilp_write_block_and_checksum(uint8_t* block, uint16_t block_length) {
	enum tilp_result_t result = TILP_OK;
	uint16_t checksum = 0;
	while (block_length-- && result == TILP_OK) {
		result = tilp_write_byte(*block);
		checksum += *block++;
	}
	if (result == TILP_OK && (result = tilp_write_byte(checksum)) == TILP_OK) {
		result = tilp_write_byte(checksum >> 8);
	}
	return result;
}

// Write a block of values from ROM using the TI link protocol.
enum tilp_result_t tilp_write_block_P(PGM_VOID_P block, uint16_t block_length) {
	enum tilp_result_t result = TILP_OK;
	while (block_length-- && result == TILP_OK) {
		result = tilp_write_byte(pgm_read_byte(block++));
	}
	return result;	
}

// Read a block of values using the TI link protocol.
enum tilp_result_t tilp_read_block(uint8_t* block, uint16_t block_length) {
	enum tilp_result_t result = TILP_OK;
	while (block_length-- && result == TILP_OK) {
		result = tilp_read_byte(block++);
	}
	return result;
}

// Read a block of values and a checksum using the TI link protocol.
enum tilp_result_t tilp_read_block_and_checksum(uint8_t* block, uint16_t block_length) {
	enum tilp_result_t result = tilp_read_block(block, block_length);
	uint8_t value;
	if (result == TILP_OK && (result = tilp_read_byte(&value)) == TILP_OK) {
		result = tilp_read_byte(&value);
	}
	return result;
}

// Writes an acknowledgement.
enum tilp_result_t tilp_write_acknowledgement(uint8_t sender) {
	uint8_t packet[4] = { sender, 0x56, 0x00, 0x00 };
	return tilp_write_block(packet, 4);
}

// Reads an acknowledgement.
enum tilp_result_t tilp_read_acknowledgement(uint8_t* sender) {
	enum tilp_result_t result = TILP_OK;
	uint8_t packet[4];
	if ((result = tilp_read_block(packet, 4)) != TILP_OK) {
		return result;
	} else {
		*sender = packet[0];
		if (packet[1] != 0x56) {
			result = TILP_NOT_ACKNOWLEDGED;
		}
		return result;
	}
}