#include <avr/io.h>
#include "video.h"

; Registers.
#define r_pixel_row    r15 /* Stores a row of pixels from the font data temporarily. */
#define r_column_count r16 /* Counts down columns. */

; Time constants.

; PAL timing constants (microseconds).
#define pal_t_short_vsync          4 /* Length of the "short" sync pulses. */
#define pal_t_long_vsync          28 /* Length of the "long" sync pulses. */
#define pal_t_active_sync          4 /* Length of the sync pulse on active display. */
#define pal_t_active_picture      60 /* Length of the picture data on active display. */

; NTSC timing constants (tenths of microseconds).
#define ntsc_t_vsync_black        47 /* Length of the black signal at the start of vsync lines. */
#define ntsc_t_vsync_sync        588 /* Length of the sync signal during vsync lines. */
#define ntsc_t_active_sync        47 /* Length of the sync pulse on active display. */
#define ntsc_t_active_display    588 /* Length of the active display. */

; Variables.
.lcomm timer1_compa_handler, 2     ; Function pointer to current TIMER1_COMPA_vect handler.
.lcomm scan_region_repeat, 2       ; Used to count down display regions when rasterising.
.lcomm scanline_counter, 2         ; Used to count scanlines.
.lcomm scanline_stretch_counter, 1 ; Used to count scanlines to stretch source pixel rows vertically.
.lcomm scanline_stretch_number, 1  ; Used to index scanlines to stretch source pixel rows vertically.

#define cycles_us(us) (us)*20-1
#define cycles_tenth_us(tus) (tus)*2-1

.global video_driver_init_pal
video_driver_init_pal:
	push r20
	ldi r20, lo8(pm(pal_long_vsync_sync))
	sts timer1_compa_handler+0, r20
	ldi r20, hi8(pm(pal_long_vsync_sync))
	sts timer1_compa_handler+1, r20
	ldi r20, 5
	sts scan_region_repeat, r20
	pop r20
	ret

.global video_driver_init_ntsc
video_driver_init_ntsc:
	push r20
	ldi r20, lo8(pm(ntsc_vsync_black))
	sts timer1_compa_handler+0, r20
	ldi r20, hi8(pm(ntsc_vsync_black))
	sts timer1_compa_handler+1, r20
	ldi r20, 10
	sts scan_region_repeat, r20
	pop r20
	ret

.global TIMER1_COMPA_vect
TIMER1_COMPA_vect:
	in r_sreg_save,_SFR_IO_ADDR(SREG) ; [1]
	push ZL                           ; [2]
	push ZH                           ; [2]
	push YL                           ; [2]
	push YH                           ; [2]

	; Disable CTC.
	ldi ZL, _BV(CS10)
	sts _SFR_MEM_ADDR(TCCR1B),ZL
	
	lds ZL,_SFR_MEM_ADDR(TCNT1L) ; [2]
	
	sbrs ZL,0                    ; [2/1] \_ bit 0 = 0: 3cc \  00 -> 7
	rjmp .                       ; [2]   /  bit 0 = 1: 2cc  | 01 -> 6
	sbrs ZL,1                    ; [2/1] \_ bit 1 = 0: 4cc  | 10 -> 5
	lpm ZL,Z                     ; [3]   /  bit 1 = 1: 2cc /  11 -> 4

	lds ZL, timer1_compa_handler+0      ; [2]
	lds ZH, timer1_compa_handler+1      ; [2]
	ijmp                                ; [2]

; Exits the interrupt handler.
exit_timer1_compa_vect:
	sts timer1_compa_handler+0, ZL
	sts timer1_compa_handler+1, ZH
	sts _SFR_MEM_ADDR(OCR1AH), YH
	sts _SFR_MEM_ADDR(OCR1AL), YL
exit_timer1_compa_vect_no_store:
	; Re-enable CTC.
	ldi ZL, _BV(WGM12) | _BV(CS10)
	sts _SFR_MEM_ADDR(TCCR1B),ZL

	; Clear interrupts (if any happened).
	sbi _SFR_IO_ADDR(TIFR1), OCF1A

	pop YH
	pop YL
	pop ZH
	pop ZL
	out _SFR_IO_ADDR(SREG),r_sreg_save ; [1]
	reti

; 30uS of sync.
pal_long_vsync_sync:
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT
	ldi YL,lo8(cycles_us(pal_t_long_vsync))
	ldi YH,hi8(cycles_us(pal_t_long_vsync))
	ldi ZL,lo8(pm(pal_long_vsync_black))
	ldi ZH,hi8(pm(pal_long_vsync_black))
	jmp exit_timer1_compa_vect

; 2uS of black.
pal_long_vsync_black:
	sbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT
	; Is this the last long vsync?
	lds YL, scan_region_repeat
	dec YL
	brne pal_long_vsync_not_finished
	; Long sync is always followed by 5x short sync.
	ser YL
	sts scan_region_repeat+1, YL ; Used to denote that we're entering short sync from long sync.
	ldi YL, 5
	ldi ZL,lo8(pm(pal_short_vsync_sync))
	ldi ZH,hi8(pm(pal_short_vsync_sync))
	jmp pal_long_vsync_finished
pal_long_vsync_not_finished:
	ldi ZL,lo8(pm(pal_long_vsync_sync))
	ldi ZH,hi8(pm(pal_long_vsync_sync))
pal_long_vsync_finished:
	sts scan_region_repeat, YL
	ldi YL,lo8(cycles_us(pal_t_short_vsync))
	ldi YH,hi8(cycles_us(pal_t_short_vsync))
	jmp exit_timer1_compa_vect

; 2uS of sync.
pal_short_vsync_sync:
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT
	cbi _SFR_IO_ADDR(VIDEO_PICTURE_PORT),VIDEO_PICTURE_BIT
	ldi YL,lo8(cycles_us(pal_t_short_vsync))
	ldi YH,hi8(cycles_us(pal_t_short_vsync))
	ldi ZL,lo8(pm(pal_short_vsync_black))
	ldi ZH,hi8(pm(pal_short_vsync_black))
	jmp exit_timer1_compa_vect

; 30uS of black.
pal_short_vsync_black:
	sbi _SFR_IO_ADDR(VIDEO_SYNC_PORT), VIDEO_SYNC_BIT
	; Is this the last short vsync?
	lds YL,scan_region_repeat
	dec YL
	brne pal_short_vsync_not_finished
	; Short sync is either followed by 5x long sync or active display.
	lds YL,scan_region_repeat+1
	inc YL
	breq pal_short_vsync_enter_pal_active_display
pal_short_vsync_enter_long_sync:
	; We've just finished a frame!		
	ldi YL, 5
	ldi ZL,lo8(pm(pal_long_vsync_sync))
	ldi ZH,hi8(pm(pal_long_vsync_sync))
	jmp pal_short_vsync_finished
pal_short_vsync_enter_pal_active_display:
	; We're about to enter active display.
	; Initialise the scanline counter, skipping the first few lines.
	lds YL,video_pal_y_offset+0
	lds YH,video_pal_y_offset+1
	sts scanline_counter+0,YL
	sts scanline_counter+1,YH
	; 304 lines in PAL.
	ldi YH,hi8(-304)
	sts scan_region_repeat+1, YH
	ldi YL,lo8(-304)
	ldi ZL,lo8(pm(pal_active_sync))
	ldi ZH,hi8(pm(pal_active_sync))
	jmp pal_short_vsync_finished
pal_short_vsync_not_finished:
	ldi ZL,lo8(pm(pal_short_vsync_sync))
	ldi ZH,hi8(pm(pal_short_vsync_sync))
pal_short_vsync_finished:
	sts scan_region_repeat, YL
	ldi YL,lo8(cycles_us(pal_t_long_vsync))
	ldi YH,hi8(cycles_us(pal_t_long_vsync))
	jmp exit_timer1_compa_vect

; 4uS of sync.
pal_active_sync:
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT
	cbi _SFR_IO_ADDR(VIDEO_PICTURE_PORT),VIDEO_PICTURE_BIT

	; Increment the scanline counter. [10]
	lds YL,scanline_counter+0 ; [2]
	lds YH,scanline_counter+1 ; [2]
	adiw YL,1                 ; [1]
	sts scanline_counter+0,YL ; [2]
	sts scanline_counter+1,YH ; [2]

	ldi YL,lo8(cycles_us(pal_t_active_sync))
	ldi YH,hi8(cycles_us(pal_t_active_sync))
	ldi ZL,lo8(pm(pal_active_back_porch))
	ldi ZH,hi8(pm(pal_active_back_porch))
	jmp exit_timer1_compa_vect

; 8uS of back porch.
pal_active_back_porch:
	sbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]

	; Have we finished scanning the active display? [25]
	lds YL,scan_region_repeat+0              ; [2]
	lds YH,scan_region_repeat+1              ; [2]
	adiw YL,1                                ; [2]
	sts scan_region_repeat+0,YL              ; [2]
	sts scan_region_repeat+1,YH              ; [2]
	brne pal_pal_active_display_not_finished ; [2/1]

pal_pal_active_display_finished:
	ldi ZL,lo8(pm(pal_short_vsync_sync))    ; [1] \.
	ldi ZH,hi8(pm(pal_short_vsync_sync))    ; [1]  |
	ldi YL,6                                ; [1]  |_ [7]
	sts scan_region_repeat,YL               ; [2]  |
	jmp pal_pal_active_display_exit         ; [2] /'
pal_pal_active_display_not_finished_delay:
	rjmp .                                  ; [2] \.
	rjmp .                                  ; [2]  |_ [6]
	rjmp .                                  ; [2] /'
pal_pal_active_display_not_finished:
	ldi ZL,lo8(pm(pal_active_sync))         ; [1] \_ [2]
	ldi ZH,hi8(pm(pal_active_sync))         ; [1] /

	rjmp .                                  ; [2] \.
	rjmp .                                  ; [2]  |_ [5]
	nop                                     ; [1] /'

pal_pal_active_display_exit:
	ldi YL,lo8(cycles_us(pal_t_active_picture)) ; [1] \_ [2]
	ldi YH,hi8(cycles_us(pal_t_active_picture)) ; [1] /

	sts timer1_compa_handler+0,ZL           ; [2] \.
	sts timer1_compa_handler+1,ZH           ; [2]  |_ [8]
	sts _SFR_MEM_ADDR(OCR1AH),YH            ; [2]  |
	sts _SFR_MEM_ADDR(OCR1AL),YL            ; [2] /'

	push r0
	push r1
	push XL
	push XH

pal_active_display:

	; More porch delay.
	lds XL,video_pal_x_offset
1:	dec XL
	brne 1b

	ldi XL,0xFF
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),XL
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]

	; Are we on the screen?
	lds XL,scanline_counter+1
	or XL,XL
	breq 1f
	jmp pal_active_display_skip_scanline
1:

	; Slight delay before sending anything to the screen.
	ldi XL,20
1:	dec XL
	brne 1b

	; Preserve registers.
	push r_pixel_row
	push r_column_count

	; Set up X to point at the current video buffer scanline.
	
	lds r0,scanline_counter+0
	lsr r0
	lsr r0
	ldi XL,12
	mul r0,XL
	movw XL,r0
	
	ldi YL,lo8(video_buffer)
	ldi YH,hi8(video_buffer)
	
	add XL,YL
	adc XH,YH

	; We have twelve columns.
	ldi r_column_count,12

pal_column_loop:

	; (1)
	ld r_pixel_row,X+                                ; [2]
	com r_pixel_row                                  ; [1]
	sbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	nop                                              ; [1]
	rjmp .                                           ; [2]

	; (2)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	nop                                              ; [1]
	rjmp .                                           ; [2]

	; (3)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	nop                                              ; [1]
	rjmp .                                           ; [2]

	; (4)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	nop                                              ; [1]
	rjmp .                                           ; [2]
	
	; (5)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	nop                                              ; [1]
	rjmp .                                           ; [2]

	; (6)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	nop                                              ; [1]
	rjmp .                                           ; [2]

	; (7)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	nop                                              ; [1]
	rjmp .                                           ; [2]

	; (8)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	dec r_column_count                               ; [1]
	nop                                              ; [1]
	brne pal_column_loop                             ; [2/1]
	nop                                              ; [1]
	
	; Finish with white.
	ldi XL,0xFF                                      ; [1]
	rjmp .                                           ; [2]
	nop                                              ; [1]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),XL          ; [1]
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]

	; Restore registers.
	pop r_column_count
	pop r_pixel_row

pal_active_display_skip_scanline:
	pop XH
	pop XL
	pop r1
	pop r0
	jmp exit_timer1_compa_vect_no_store


; 4.7uS of black.
ntsc_vsync_black:
	sbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT
	ldi YL,lo8(cycles_tenth_us(ntsc_t_vsync_black))
	ldi YH,hi8(cycles_tenth_us(ntsc_t_vsync_black))
	ldi ZL,lo8(pm(ntsc_vsync_sync))
	ldi ZH,hi8(pm(ntsc_vsync_sync))
	jmp exit_timer1_compa_vect

; 58.8uS of sync.
ntsc_vsync_sync:
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT
	; Is this the vsync line?
	lds YL, scan_region_repeat
	dec YL
	brne ntsc_vsync_not_finished

	; We're about to enter active display.
	; Initialise the scanline counter, skipping the first few lines.
	lds YL,video_ntsc_y_offset+0
	lds YH,video_ntsc_y_offset+1
	sts scanline_counter+0,YL
	sts scanline_counter+1,YH
	
	clr YL
	sts scanline_stretch_number,YL
	ldi YL,3
	sts scanline_stretch_counter,YL

	; 242 lines in NTSC.
	ldi YH,hi8(-252)
	sts scan_region_repeat+1, YH
	ldi YL,lo8(-252)
	ldi ZL,lo8(pm(ntsc_active_sync))
	ldi ZH,hi8(pm(ntsc_active_sync))
	jmp ntsc_vsync_finished
	
ntsc_vsync_not_finished:
	ldi ZL,lo8(pm(ntsc_vsync_black))
	ldi ZH,hi8(pm(ntsc_vsync_black))
ntsc_vsync_finished:
	sts scan_region_repeat, YL
	ldi YL,lo8(cycles_tenth_us(ntsc_t_vsync_sync))
	ldi YH,hi8(cycles_tenth_us(ntsc_t_vsync_sync))
	jmp exit_timer1_compa_vect

; 4.7uS of sync.
ntsc_active_sync:
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT
	cbi _SFR_IO_ADDR(VIDEO_PICTURE_PORT),VIDEO_PICTURE_BIT
	
	; Increment the scanline counter. [10]
	lds YL,scanline_counter+0 ; [2]
	lds YH,scanline_counter+1 ; [2]
	adiw YL,1                 ; [1]
	sts scanline_counter+0,YL ; [2]
	sts scanline_counter+1,YH ; [2]

	ldi YL,lo8(cycles_tenth_us(ntsc_t_active_sync))
	ldi YH,hi8(cycles_tenth_us(ntsc_t_active_sync))
	ldi ZL,lo8(pm(ntsc_active_display))
	ldi ZH,hi8(pm(ntsc_active_display))
	jmp exit_timer1_compa_vect

; 58.8uS of active display.
ntsc_active_display:
	sbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]

	; Have we finished scanning the active display? [25]
	lds YL,scan_region_repeat+0            ; [2]
	lds YH,scan_region_repeat+1            ; [2]
	adiw YL,1                              ; [2]
	sts scan_region_repeat+0,YL            ; [2]
	sts scan_region_repeat+1,YH            ; [2]
	brne ntsc_active_display_not_finished  ; [2/1]

	ldi YL,10 ; 20 vsync lines.
	sts scan_region_repeat+0,YL
	ldi YL,lo8(cycles_tenth_us(ntsc_t_active_display))
	ldi YH,hi8(cycles_tenth_us(ntsc_t_active_display))
	ldi ZL,lo8(pm(ntsc_vsync_black))
	ldi ZH,hi8(pm(ntsc_vsync_black))
	jmp exit_timer1_compa_vect

ntsc_active_display_not_finished:

	push XL
	push XH
	push r0
	push r1

	; More porch delay.
	lds XL,video_ntsc_x_offset
1:	dec XL
	brne 1b

	ldi XL,0xFF
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),XL
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]

	; Are we on the screen?
	lds XL,scanline_counter+1
	or XL,XL
	breq 1f
	jmp ntsc_active_display_skip_scanline
1:

	lds XL,scanline_stretch_number
	cpi XL,64
	brcs 1f
	jmp ntsc_active_display_skip_scanline
1:

	; Slight delay before sending anything to the screen.
	ldi XL,28
1:	dec XL
	brne 1b

	; Preserve registers.
	push r_pixel_row
	push r_column_count

	; Set up X to point at the current video buffer scanline.
	
	lds r0,scanline_stretch_number
	ldi XL,12
	mul r0,XL
	movw XL,r0
	
	ldi YL,lo8(video_buffer)
	ldi YH,hi8(video_buffer)
	
	add XL,YL
	adc XH,YH

	; We have twelve columns.
	ldi r_column_count,12

ntsc_column_loop:

	; (1)
	ld r_pixel_row,X+                                ; [2]
	com r_pixel_row                                  ; [1]
	sbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	rjmp .                                           ; [2]

	; (2)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	rjmp .                                           ; [2]

	; (3)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	rjmp .                                           ; [2]

	; (4)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	rjmp .                                           ; [2]
	
	; (5)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	rjmp .                                           ; [2]

	; (6)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	rjmp .                                           ; [2]

	; (7)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	rol r_pixel_row                                  ; [1]
	rjmp .                                           ; [2]

	; (8)
	rjmp .                                           ; [2]
	rjmp .                                           ; [2]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),r_pixel_row ; [1]
	dec r_column_count                               ; [1]
	brne ntsc_column_loop                            ; [2/1]
	nop                                              ; [1]
	
	; Finish with white.
	ldi XL,0xFF                                      ; [1]
	rjmp .                                           ; [2]
	nop                                              ; [1]
	out _SFR_IO_ADDR(VIDEO_PICTURE_PORT),XL          ; [1]
	cbi _SFR_IO_ADDR(VIDEO_SYNC_PORT),VIDEO_SYNC_BIT ; [1]

	; Do we need to advance to the next stretched row number?
	lds YL,scanline_stretch_counter
	dec YL
	brne 1f
	lds YL,scanline_stretch_number
	inc YL
	sts scanline_stretch_number,YL
	ldi YL,3
1:	sts scanline_stretch_counter,YL
	

	; Restore registers.
	pop r_column_count
	pop r_pixel_row

ntsc_active_display_skip_scanline:

	pop r1
	pop r0
	pop XH
	pop XL

	ldi YL,lo8(cycles_tenth_us(ntsc_t_active_display))
	ldi YH,hi8(cycles_tenth_us(ntsc_t_active_display))
	ldi ZL,lo8(pm(ntsc_active_sync))
	ldi ZH,hi8(pm(ntsc_active_sync))
	
	jmp exit_timer1_compa_vect